#include <stdio.h>
#include <time.h>

// crank includes
#include "common_trigram.h"
#include "common_monoalphabetic.h"
#include "files.h"

// ga specific includes
#include <ga/ga.h>
#include <ga/GAStringGenome.h>
#include <ga/GAStringGenome.C>

// This is the declaration of the objective function
float objective(GAGenome &);
// Here we declare our own initializer
void AlphabetInitializer(GAGenome &);


double *slft_std;    /* Standard unigram table */
double *bift_std;    /* Standard bigram table */
double *trift_std;   /* Standard trigram table */

char *text;
char k[(int)('Z'+1)];

int main (int argc, char *argv[]){

	int popsize	= 25;
	int ngen	= 1000;
	float pcross	= 0.9;
	float pmut	= 0.01;

	// these are used by ngen=0
	float score     = 0;
	int gennum	= 0;
	int genstep	= 100;
	int maxgen	= 0;
	

	for(int ii=1; ii<argc; ii++) {
      		if(strcmp(argv[ii],"ps") == 0) {
            		popsize = atoi(argv[++ii]); 
	        }
      		if(strcmp(argv[ii],"ng") == 0) {
            		ngen = atoi(argv[++ii]); 
	        }
      		if(strcmp(argv[ii],"pm") == 0) {
            		pmut = atof(argv[++ii]); 
	        }
      		if(strcmp(argv[ii],"pc") == 0) {
            		pcross = atof(argv[++ii]); 
	        }
      		if(strcmp(argv[ii],"mg") == 0) {
            		maxgen = atoi(argv[++ii]); 
	        }
	}
		  
	printf ("Running GA with the following parameters: ps=%02d ng=%04d mg=%04d pm=%.2f pc=%.2f\n",popsize,ngen,maxgen,pmut,pcross);

	int i;
	time_t time1,time2;
        time(&time1);
	
	/* Load 3-gram tables */
	slft_std = load_slft("slft.dat");
        if (!slft_std) return 0;
	bift_std = load_bift("bift.dat");
	if (!bift_std) return 0;
	trift_std = load_trift("trift.dat");
	if (!trift_std) return 0;

	/* Load ciphered file */
	text = file_load_text ("zorxv.txt");

	// Load parameters
	GAParameterList params;
	GASteadyStateGA::registerDefaultParameters(params); // We'll use GASteadyStateGA
	params.set(gaNpopulationSize, popsize);    // population size
	params.set(gaNpCrossover, pcross);       // probability of crossover 
	params.set(gaNpMutation, pmut);       // probability of mutation
	params.set(gaNnGenerations, ngen);    // number of generations
	params.parse(argc, argv, gaFalse);

	// Create GAStringAlleleSet
	GAStringAlleleSet alleles;
	for(i=0; i<26; i++)
		alleles.add('A'+i); // alleles are letters from A to Z
	
	GAStringGenome genome(26, alleles, objective);	// passes 1) Size 2) AlleleSet 3) Objective
	genome.initializer(AlphabetInitializer);	// uhmm... puo' andare, basta che sian tutti diversi
	genome.mutator(GAStringSwapMutator);		// ci vuole per mantenere l'unicita' delle lettere
	GASteadyStateGA ga(genome);			// va bene anche il simple?
	ga.parameters(params);
	ga.crossover(GAStringPartialMatchCrossover);	// anche questo ci vuole
	
	if (ngen!=0){
		ga.evolve();
	}else{
		ga.initialize();
		//while (score<1067){
		while (score<8.764){
			for (int i=0;i<genstep;i++) ga.step();
			score = ga.statistics().bestIndividual().score();
			printf ("Generation: %04d Score: %04.04f\n",gennum,score);
			gennum+=genstep;
			if (maxgen && gennum>=maxgen) break;
		}
	}
	
	printf ("Generations: %d ",gennum);
	if (maxgen && gennum>=maxgen){
		printf("- maxgen reached (%d), GA stoppeth",maxgen);
	}
	printf ("\n");
	
	time(&time2);
	genome = ga.statistics().bestIndividual();
	cout << "the ga generated the following string (objective score is ";
	cout << genome.score() << "):\n" << genome << "\n";
	printf ("Total time: %d seconds\n",(int)(time2-time1));
	
	printf ("%s\n",argv[0]);
	
	return 0;

}

void genome2key (GAGenome & c, key *k){
	// genome to key conversion, needed inside objective function
	GAStringGenome & genome = (GAStringGenome &)c;
	for(int i=0; i<genome.size(); i++){
		(*k)['A'+i]=genome.gene(i);
	}
	
}

float objective(GAGenome & c){
	GAStringGenome & genome = (GAStringGenome &)c;

	int len;
	float score = 0;
	double *slft, *bift, *trift;
	double slft_err, bift_err, trift_err;
	char *decoded_text;
	char k[(int)('Z'+1)];

	// convert the genome into a key
	genome2key (genome,&k);
	// apply the key to the text to generate a plaintext candidate
	decoded_text = apply_key_text (&k,text);
	// evaluate the candidate calculating frequency table
	len = make_ft(decoded_text, &slft, &bift, &trift);

	slft_err = slft_error (slft_std, slft);
	bift_err = bift_error (bift_std, bift);
	trift_err = trift_error (trift_std, trift);

	free (decoded_text);
	free (slft);
	free (bift);
	free (trift);

	score = 1/(slft_err+10*bift_err+100*trift_err); // 8.764
	//score = 1/(trift_err); // 1067.21
	return score;
}

void AlphabetInitializer(GAGenome & c){
	GAStringGenome &genome=(GAStringGenome &)c;
	int i;
	for(i=0; i<genome.size(); i++)
		genome.gene(25-i, 'a'+i);
	
	for(i=0; i<genome.size(); i++)
		if(GARandomBit()) genome.swap(i, GARandomInt(0, genome.size()-1));
}


